"""
Classes that define the file format for pecha files (projects).
"""
import os
import shutil
import tempfile
from zipfile import ZipFile

import sqlalchemy

from pechapub import error
from pechapub.datamodel import schema


EXT=".ppf"
DB_NAME = "data.sqlite"
TIBETAN_NAME = "tibetan.rtf"
TRANSLITER_NAME = "transliter.rtf"
TRANSLAT_NAME = "translat.rft"

TEXT_FILES = [
    TIBETAN_NAME,
    TRANSLITER_NAME,
    TRANSLAT_NAME,
]

BACKUP_NAME = "backup"
BACKUP_PROJECT_NAME = "backup/project"
BACKUP_FILES_NAME = "backup/files"
CONTENT_NAME = "content"
CONTENT_IMAGES_NAME = "content/images"
CONTENT_TEXT_NAME = "content/text"
FONTS_NAME = "fonts"

DIRECTORY_STRUCTURE = [
    BACKUP_NAME,
    BACKUP_PROJECT_NAME,
    BACKUP_FILES_NAME,
    CONTENT_NAME,
    CONTENT_IMAGES_NAME,
    CONTENT_TEXT_NAME,
    FONTS_NAME,
]

DEFAULT_PROJECT_BACKUP_COUNT = 5
DEFAULT_FILE_BACKUP_COUNT = 20


class ProjectAspect(object):

    def __init__(self, parent):
        self.parent = parent


class ProjectDatabase(ProjectAspect):

    def __init__(self, parent, create=False):
        super(ProjectDatabase, self).__init__(parent)
        self.path = os.path.join(
            self.parent.base_dir, self.parent.project_name, DB_NAME)
        if create:
            self._create_database()

    def _create_database(self):
        self.engine = sqlalchemy.create_engine("sqlite:///%s" % self.path)
        # XXX do this in a better way...
        metadata = sqlalchemy.MetaData()
        schema.create_schema(metadata)
        metadata.create_all(self.engine)
        # XXX bind the db session and set as a class attribute

    # XXX There will need to be convenience methods here for performing the
    # basic data read and write operations. This will make the GUI programming
    # much easier.

class ProjectTexts(ProjectAspect):

    def __init__(self, parent, create=False):
        super(ProjectTexts, self).__init__(parent)
        self.path = os.path.join(
            self.parent.base_dir, self.parent.project_name, CONTENT_TEXT_NAME)
        if create:
            self._create_text_files()

    def _create_text_files(self):
        for file_name in TEXT_FILES:
            fh = open(os.path.join(self.path, file_name), "w")
            fh.write("")
            fh.close()

    def _read(self, file_type):
        file_path = os.path.join(self.path, file_type)
        return open(file_path).read()

    def _write(self, file_type, text_data):
        file_path = os.path.join(self.path, file_type)
        open(file_path, "w").write(text_data)

    def read_tibetan(self):
        return self._read(TIBETAN_NAME)

    def write_tibetan(self, text):
        self._write(TIBETAN_NAME, text)

    def read_transliteration(self):
        return self._read(TRANSLITER_NAME)

    def write_transliteration(self, text):
        self._write(TRANSLITER_NAME, text)

    def read_translation(self):
        return self._read(TRANSLAT_NAME)

    def write_translation(self, text):
        self._write(TRANSLAT_NAME, text)


# XXX needs implementation and unit tests!
class ProjectImages(ProjectAspect):

    def read_image(self, image_name):
        pass

    def write_image(self, image_name, data):
        pass


class BaseProject(object):

    def __init__(self, base_dir="", project_name=""):
        self.base_dir = base_dir
        if project_name:
            project_name = self._get_project_name(project_name)
        self.project_name = project_name
        self.project_path = self._get_project_path()
        # placeholders for aspects
        self.db = None
        self.texts = None
        self.images = None

    def _get_project_name(self, project_name):
        if not project_name.endswith(".ppf"):
            project_name += EXT
        return project_name

    def _get_project_path(self):
        return os.path.join(self.base_dir, self.project_name)


class Project(BaseProject):
    """

    """
    def create(self, base_dir="", project_name=""):
        self._set_attributes(base_dir, project_name)
        if os.path.exists(self.project_path):
            raise error.ProjectCreationError(
                "A project with that name already exists in the given "
                "directory. Please choose another name.")
        self._create_directories()
        self._set_aspects(create=True)

    def _set_attributes(self, base_dir="", project_name=""):
        if base_dir:
            self.base_dir = base_dir
        if project_name:
            self.project_name = self._get_project_name(project_name)
        if base_dir or project_name:
            self.project_path = self._get_project_path()

    def _create_directories(self):
        for dir_name in DIRECTORY_STRUCTURE:
            path = os.path.join(self.project_path, dir_name)
            os.makedirs(path)

    def _set_aspects(self, create=False):
        self.db = ProjectDatabase(self, create)
        self.texts = ProjectTexts(self, create)
        self.images = ProjectImages(self)

    def load(self, project_path):
        """
        When a project is loaded from a directory on the file system, the
        following needs to occur:
            * base_dir, project_name, and project_path need to get set
            * the db, texts, and images aspects need to be set
        """
        if not project_path.endswith(EXT):
            raise error.ProjectLoadError(
                "PechaPub does not recognize projects that do not have the "
                "%s extension." % EXT)
        if not os.path.isdir(project_path):
            raise error.ProjectLoadError(
                "To load an uncompressed PechaPub project, the project must "
                "be a directory and it must exist.")
        base_dir = os.path.dirname(project_path)
        project_name = os.path.basename(project_path)
        self._set_attributes(base_dir, project_name)
        self._set_aspects()


class CompressedProject(object):
    """
    This class links a project directory with a zip file, and vice versa. If a
    compressed file is provided, a temporary directory is created where
    extracted contents may be accessed and written to. If a project object is
    provided, it is converted to a zip archive. If a project directory is
    provided, it's is converted to a project object and then to a zip archive.
    If the project directory doesn't exist, a new one is created.
    """
    # XXX hrm... it seems the only thing missing here is the final project
    # object... it should not be the source dir, but rather the temp dir
    # created from the extracted zip. Source dirs need to be converted to zips
    # and removed (backed up). Every instance's .project attribute should point
    # to the extracted zip temp directory.
    def __init__(self, project_object=None, project_path=""):
        if project_object and not isinstance(project_object, BaseProject):
            raise ValueError(
                "The project object must be a subclass of BaseProject.")
        # if a project object is not provided, let's see if we can create one
        if not project_object:
            if not project_path:
                raise error.ProjectCreationError(
                    "You must provide either a project object or the path to "
                    "a project")
            project_object = self._get_project_object(project_path)
        self.zip = self.convert(project_object)
        self.project = project_object
        extract_base = tempfile.mkdtemp()
        self.extract_path = os.path.join(
            tempfile.mkdtemp(), project_object.project_name)

    def _get_project_object(self, project_path):
        """
        Give a project path, a project object is returned. This covers the
        following cases:
            * the provided path is a directory and it exists
            * the provided directory does not exist

        XXX This method does nothing if the project_path points to a zip file.
        (should it?)
        """
        base_dir = os.path.dirname(project_path)
        project_name = os.path.basename(project_path)
        if not os.path.isfile(project_path):
            project_object = Project(base_dir, project_name)
            # if there's no project, let's make one
            if not os.path.exists(project_path):
                project_object.create()
            # if there's a project there, then let's load it
            elif os.path.isdir(project_path):
                project_object.load(project_path) 
        return project_object

    def prepare_temp_dir(self):
        pass

    def convert(self, project_object):
        """
        This method is intended to convert a file system projct to a compressed
        project.
        """
        # backup current project

        # create temp dir

        # move current project to temp dir

        # compress moved project to old project dir location, using the old
        # projects dir name as the compressed file name

        # remove temp dir

        # return new project object for compressed project

    def backup(self, source):
        """
        Before edits, copies moves, etc., are performed on a project, we need
        to back it up. The source to be backed up can be a directory or a zip
        file.
        """
        # XXX a backup method needs to be created on the Project class, too

        # create a temporary directory where a backup can be saved

        # get the current datetime stamp, millisecond resolution

        # copy the zip file to the temporary directory

        # extract the zip file (???)
        # XXX this would be much easier with an async app... split this into
        # events. The backup needs to be performed in stages - one before the
        # opening/extracting, and then the second half once the extracting has
        # completed. Maybe we can split this method into two private methods...
        # _backup_stage1, _backup_stage2?
        # XXX and another question: where does the extracting take place?
        # always in on spot in the code? makes it easier to define when/where
        # backups occur... if not, we could be in for some more deep
        # thinking...

        # move the copied zip from from the temporary directory to the
        # ./backup/project directory

        # remove the temporary directory

    def save(self):
        """
        Bundle the contents of the working directory into a zip file and
        replace the old zip file with the new one.
        """
        #self.zip = ZipFile(path, "w")
        pass

    def _prepare_dir_project(self):
        # create a new, temporary directory
        temp_dir = tempfile.mkdtemp()
        # move the directory to a temporary location
        shutil.move(self.project_path)
        return os.path.join(temp_dir, self.project_name)

    def extract(self):
        pass
