# coding=utf-8
import os
import shutil

from pechapub import error, project
from pechapub.testing.base import BaseTestCase


class BaseProjectTestCase(BaseTestCase):

    def setUp(self):
        super(BaseProjectTestCase, self).setUp()
        self.base_dir = self.make_temp_dir()
        self.ppf = project.Project()
        self.ppf.create(self.base_dir, "MyPecha")
        self.project_path = os.path.join(self.base_dir, "MyPecha.ppf")


class ProjectAspectTestCase(BaseProjectTestCase):

    def test_base_init(self):
        aspect = project.ProjectAspect(self.ppf)
        self.assertEquals(aspect.parent.project_name, "MyPecha.ppf")


class ProjectDatabaseTestCase(ProjectAspectTestCase):

    def test_init_without_create_flag(self):
        """
        When we init without setting the create flag to true, no files should
        exist.
        """
        ppf = project.Project()
        aspect = project.ProjectDatabase(ppf)
        self.assertFalse(os.path.exists(aspect.path))

    def test_init_with_create_flag(self):
        """
        When using the project instance created in BaseProjectTestCase's setUp,
        the create method called implicitly sets the create flag to True in the
        various aspect classes, so the files should exist.

        This also implicitly tests the private _create_database method.
        """
        aspect = self.ppf.db
        self.assertTrue(os.path.exists(aspect.path))


class ProjectTextsTestCase(ProjectAspectTestCase):

    def test_init_without_create_flag(self):
        """
        When we init without setting the create flag to true, no files should
        exist.
        """
        ppf = project.Project()
        aspect = project.ProjectTexts(ppf)
        for file_name in project.TEXT_FILES:
            path = os.path.join(
                aspect.path, file_name)
            self.assertFalse(os.path.exists(path))

    def test_init_with_create_flag(self):
        """
        When using the project instance created in BaseProjectTestCase's setUp,
        the create method called implicitly sets the create flag to True in the
        various aspect classes, so the files should exist.

        This also implicitly tests the private _create_text_files method.
        """
        for file_name in project.TEXT_FILES:
            path = os.path.join(
                self.ppf.texts.path, file_name)
            self.assertTrue(os.path.exists(path))

    def test_read_and_write_tibetan(self):
        before = self.ppf.texts.read_tibetan()
        self.assertEquals(before, "")
        tibetan = "ཕྱག་འཚལ་བ་དང་མཆོད་ཅིང་བཤགས་པ་དང"
        self.ppf.texts.write_tibetan(tibetan)
        after = self.ppf.texts.read_tibetan()
        self.assertEquals(after, tibetan)

    def test_read_and_write_transliteration(self):
        before = self.ppf.texts.read_transliteration()
        self.assertEquals(before, "")
        translit = "CHAG-TSAL WA-DANG CHO-CHING SHAG-PA DANG"
        self.ppf.texts.write_transliteration(translit)
        after = self.ppf.texts.read_transliteration()
        self.assertEquals(after, translit)

    def test_read_and_write_translation(self):
        before = self.ppf.texts.read_translation()
        self.assertEquals(before, "")
        translat = "I prostrate with my body, speech and mind"
        self.ppf.texts.write_translation(translat)
        after = self.ppf.texts.read_translation()
        self.assertEquals(after, translat)


class ProjectTestCase(BaseProjectTestCase):

    def test_init(self):
        self.assertTrue(hasattr(self.ppf, "base_dir"))
        self.assertTrue(hasattr(self.ppf, "project_name"))
        self.assertTrue(hasattr(self.ppf, "project_path"))
        self.assertTrue(hasattr(self.ppf, "db"))
        self.assertTrue(hasattr(self.ppf, "texts"))
        self.assertTrue(hasattr(self.ppf, "images"))

    def test_create(self):
        # check project extension
        self.assertEquals(self.ppf.project_name, "MyPecha.ppf")
        # check directories
        for dir in project.DIRECTORY_STRUCTURE:
            path = os.path.join(self.project_path, dir)
            self.assertTrue(os.path.exists(path))

    def test_create_with_extension(self):
        ppf = project.Project()
        ppf.create(self.base_dir, "MyPecha2.ppf")
        # check project extension
        self.assertEquals(ppf.project_name, "MyPecha2.ppf")

    def test_create_when_already_created(self):
        base_dir = self.make_temp_dir()
        project.Project().create(base_dir, "Pecha1.ppf")
        ppf = project.Project()
        self.assertRaises(
            error.ProjectCreationError, ppf.create, base_dir, "Pecha1.ppf")

    def test_load(self):
        new_ppf = project.Project()
        new_ppf.load(self.ppf.project_path)
        self.assertEquals(new_ppf.base_dir, self.ppf.base_dir)
        self.assertEquals(new_ppf.project_name, self.ppf.project_name)
        self.assertEquals(new_ppf.project_path, self.ppf.project_path)
        self.assertEquals(new_ppf.db.path, self.ppf.db.path)
        self.assertEquals(new_ppf.texts.path, self.ppf.texts.path)
        
    def test_load_with_bad_extension(self):
        ppf = project.Project()
        self.assertRaises(error.ProjectLoadError, ppf.load, "/tmp/file.proj")

    def test_load_doesnt_exist(self):
        ppf = project.Project()
        self.assertRaises(
            error.ProjectLoadError, ppf.load, "/123/bogus/pecha/dir.ppf")

    def test_load_non_directory(self):
        temp_file = self.make_temp_file()
        proj_temp_file = "%s.ppf" % temp_file
        shutil.move(temp_file, proj_temp_file)
        ppf = project.Project()
        self.assertRaises(error.ProjectLoadError, ppf.load, proj_temp_file)
        # for clean-up purposes, let's move it back
        shutil.move(proj_temp_file, temp_file)


class CompressedProjectTestCase(BaseProjectTestCase):

    def test_init_bad_object(self):
        self.assertRaises(
            ValueError, project.CompressedProject, object())

    def test_init_no_params(self):
        self.assertRaises(
            error.ProjectCreationError, project.CompressedProject)

    def test_init_with_project_object(self):
        pass

    def test_init_with_non_existent_path(self):
        pass

    def test_init_with_directory_path(self):
        pass

    def test_init_with_file_parh(self):
        pass

    def test_save(self):
        ppf = project.CompressedProject(self.ppf)
        ppf.save()

    def test_extract(self):
        pass
